/*
 * $Id$
 *
 * Analog Device's SHARC 21369 compatible bus driver via BSR
 * Copyright (C) 2009- Monami Software Limited Partnership, JAPAN.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * Written by Masaki Muranaka <monaka at monami-software.com>
 *  Original version written by Ajith Kumar P.C <ajithpc@kila.com>
 *
 * Documentation:
 *      [1] Analog Devices Inc.,"ADSP-21369 SHARC Technical Reference", September 1998
 *      [2] Analog Devices Inc.,"ADSP-21369 SHARC User's Manual", September 1998
 *
 */

#include "sysdep.h"

#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include "part.h"
#include "bus.h"
#include "bssignal.h"
#include "jtag.h"
#include "buses.h"
#include "generic_bus.h"

//no SDRAM access

typedef struct {
	uint32_t last_adr;
	signal_t *addr[24];
	signal_t *data[32];
	signal_t *ms0;	//boot memory select
	signal_t *ms1;	//boot memory select
	signal_t *nwe;
	signal_t *nrd;
} bus_params_t;

#define	LAST_ADR	((bus_params_t *) bus->params)->last_adr
#define	ADDR		((bus_params_t *) bus->params)->addr
#define	DATA		((bus_params_t *) bus->params)->data
#define	MS0		((bus_params_t *) bus->params)->ms0
#define	MS1		((bus_params_t *) bus->params)->ms1
#define	nWE		((bus_params_t *) bus->params)->nwe
#define	nRD		((bus_params_t *) bus->params)->nrd

/**
 * bus->driver->(*new_bus)
 *
 */
static bus_t *sharc_21369_bus_new( chain_t *chain, const bus_driver_t *driver, char *cmd_params[] )
{
	bus_t *bus;
	part_t *part;
	char buff[15];
	int i;
	int failed = 0;

	bus = calloc( 1, sizeof (bus_t) );
	if (!bus)
		return NULL;

	bus->driver = driver;
	bus->params = calloc( 1, sizeof (bus_params_t) );
	if (!bus->params) {
		free( bus );
		return NULL;
	}

	CHAIN = chain;
	PART = part = chain->parts->parts[chain->active_part];

	for (i = 0; i < 24; i++) {
		sprintf( buff, "ADDR%d", i );
		failed |= generic_bus_attach_sig( part, &(ADDR[i]), buff );
	}

	for (i = 0; i < 32; i++) {
		sprintf( buff, "DATA%d", i );
		failed |= generic_bus_attach_sig( part, &(DATA[i]), buff );
	}

	failed |= generic_bus_attach_sig( part, &(MS0), "MS0_B" );
	failed |= generic_bus_attach_sig( part, &(MS1), "MS1_B" );
	failed |= generic_bus_attach_sig( part, &(nWE), "WR_B"  );
	failed |= generic_bus_attach_sig( part, &(nRD), "RD_B"  );

	if (failed) {
		free( bus->params );
		free( bus );
		return NULL;
	}

	return bus;
}

/**
 * bus->driver->(*printinfo)
 *
 */
static void sharc_21369_bus_printinfo( bus_t *bus )
{
	int i;

	for (i = 0; i < CHAIN->parts->len; i++)
		if (PART == CHAIN->parts->parts[i])
			break;
	printf( _("Analog Device's SHARC 21369 compatible bus driver via BSR (JTAG part No. %d)\n"), i );
}

/**
 * bus->driver->(*area)
 *
 */
static int
sharc_21369_ezkit_bus_area( bus_t *bus, uint32_t adr, bus_area_t *area )
{
	part_t *p = PART;

	if (UINT32_C(0x200000) <= adr && adr <= UINT32_C(0x027FFFF)) {

		area->description = N_("Boot Memory Select");
		area->start  = UINT32_C(0x200000);
		area->length = UINT64_C(0x080000);
		area->width = 8;

		part_set_signal( p, MS0, 1, 1);
		part_set_signal( p, MS1, 1, 0);

	} else {
		area->description = NULL;
		area->start = UINT32_C(0xffffffff);
		area->length = UINT64_C(0x080000);
		area->width = 0;

		part_set_signal( p, MS0, 1, 1);
		part_set_signal( p, MS1, 1, 1);
	}

	return URJTAG_STATUS_OK;
}

static void
setup_address( bus_t *bus, uint32_t a )
{
	int i;
	part_t *p = PART;

	for (i = 0; i < 24; i++)
		part_set_signal( p, ADDR[i], 1, (a >> i) & 1 );
}

static void
set_data_in( bus_t *bus, uint32_t adr )
{
	int i;
	part_t *p = PART;
	bus_area_t area;

	bus->driver->area( bus, adr, &area );

	for (i = 0; i < area.width; i++)
		part_set_signal( p, DATA[i], 0, 0 );
}


static void
setup_data( bus_t *bus, uint32_t adr, uint32_t d )
{
	int i;
	part_t *p = PART;
	bus_area_t area;

	bus->driver->area( bus, adr, &area );

	for (i = 0; i < area.width; i++)
		part_set_signal( p, DATA[i], 1, (d >> i) & 1 );
}

/**
 * bus->driver->(*read_start)
 *
 */
static void
sharc_21369_bus_read_start( bus_t *bus, uint32_t adr )
{
	chain_t *chain = CHAIN;
	part_t *p = PART;
	bus_area_t area;

	bus->driver->area( bus, adr, &area );

	LAST_ADR = adr;

	part_set_signal( p, nWE, 1, 1 );
	part_set_signal( p, nRD, 1, 0 );

	setup_address( bus, adr );
	set_data_in( bus, adr );

	chain_shift_data_registers( chain, 0 );
}

/**
 * bus->driver->(*read_next)
 *
 */
static uint32_t
sharc_21369_bus_read_next( bus_t *bus, uint32_t adr )
{
	part_t *p = PART;
	chain_t *chain = CHAIN;
	uint32_t d = 0;
	int i;
	bus_area_t area;

	LAST_ADR = adr;

	bus->driver->area( bus, adr, &area );

	setup_address( bus, adr );
	chain_shift_data_registers( chain, 1 );

	for (i = 0; i < area.width; i++)
		d |= (uint32_t) (part_get_signal( p, DATA[i] ) << i);

	return d;
}

/**
 * bus->driver->(*read_end)
 *
 */
static uint32_t
sharc_21369_bus_read_end( bus_t *bus )
{
	part_t *p = PART;
	chain_t *chain = CHAIN;
	int i;
	uint32_t d = 0;
	bus_area_t area;

	bus->driver->area( bus, LAST_ADR, &area );

	part_set_signal( p, nWE, 1, 1 );
	part_set_signal( p, nRD, 1, 1 );

	chain_shift_data_registers( chain, 1 );

	for (i = 0; i < area.width; i++)
		d |= (uint32_t) (part_get_signal( p, DATA[i] ) << i);

	return d;
}

/**
 * bus->driver->(*write)
 *
 */
static void
sharc_21369_bus_write( bus_t *bus, uint32_t adr, uint32_t data )
{
	part_t *p = PART;
	chain_t *chain = CHAIN;
	bus_area_t area;

	bus->driver->area( bus, adr, &area );

	part_set_signal( p, nWE, 1, 1 );
	part_set_signal( p, nRD, 1, 1 );

	setup_address( bus, adr );
	setup_data( bus, adr, data );

	chain_shift_data_registers( chain, 0 );

	part_set_signal( p, nWE, 1, 0 );
	chain_shift_data_registers( chain, 0 );
	part_set_signal( p, nWE, 1, 1 );
	chain_shift_data_registers( chain, 0 );
}

const bus_driver_t sharc_21369_ezkit_bus = {
	"SHARC_21369_EZKIT",
	N_("SHARC_21369 EZ-KIT bus driver via BSR"),
	sharc_21369_bus_new,
	generic_bus_free,
	sharc_21369_bus_printinfo,
	generic_bus_prepare_extest,
	sharc_21369_ezkit_bus_area,
	sharc_21369_bus_read_start,
	sharc_21369_bus_read_next,
	sharc_21369_bus_read_end,
	generic_bus_read,
	sharc_21369_bus_write,
	generic_bus_no_init
};
